@@ -1,21 +1,20 @@
 /* efi.h - declare EFI types and functions */
 /*
  *  GRUB  --  GRand Unified Bootloader
- *  Copyright (C) 2006  Free Software Foundation, Inc.
+ *  Copyright (C) 2006,2007,2008  Free Software Foundation, Inc.
  *
- *  This program is free software; you can redistribute it and/or modify
+ *  GRUB is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
+ *  the Free Software Foundation, either version 3 of the License, or
  *  (at your option) any later version.
  *
- *  This program is distributed in the hope that it will be useful,
+ *  GRUB is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
- *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
  */
 
 #ifndef GRUB_EFI_API_HEADER
@@ -41,15 +40,15 @@
 #define GRUB_EFI_TPL_NOTIFY		16
 #define GRUB_EFI_TPL_HIGH_LEVEL		31
 
-#define GRUB_EFI_MEMORY_UC	0x0000000000000001
-#define GRUB_EFI_MEMORY_WC	0x0000000000000002
-#define GRUB_EFI_MEMORY_WT	0x0000000000000004
-#define GRUB_EFI_MEMORY_WB	0x0000000000000008
-#define GRUB_EFI_MEMORY_UCE	0x0000000000000010
-#define GRUB_EFI_MEMORY_WP	0x0000000000001000
-#define GRUB_EFI_MEMORY_RP	0x0000000000002000
-#define GRUB_EFI_MEMORY_XP	0x0000000000004000
-#define GRUB_EFI_MEMORY_RUNTIME	0x8000000000000000
+#define GRUB_EFI_MEMORY_UC	0x0000000000000001LL
+#define GRUB_EFI_MEMORY_WC	0x0000000000000002LL
+#define GRUB_EFI_MEMORY_WT	0x0000000000000004LL
+#define GRUB_EFI_MEMORY_WB	0x0000000000000008LL
+#define GRUB_EFI_MEMORY_UCE	0x0000000000000010LL
+#define GRUB_EFI_MEMORY_WP	0x0000000000001000LL
+#define GRUB_EFI_MEMORY_RP	0x0000000000002000LL
+#define GRUB_EFI_MEMORY_XP	0x0000000000004000LL
+#define GRUB_EFI_MEMORY_RUNTIME	0x8000000000000000LL
 
 #define GRUB_EFI_OPEN_PROTOCOL_BY_HANDLE_PROTOCOL	0x00000001
 #define GRUB_EFI_OPEN_PROTOCOL_GET_PROTOCOL		0x00000002
@@ -89,6 +88,26 @@
     { 0x8e, 0x39, 0x00, 0xa0, 0xc9, 0x69, 0x72, 0x3b } \
   }
 
+#define GRUB_EFI_MPS_TABLE_GUID	\
+  { 0xeb9d2d2f, 0x2d88, 0x11d3, \
+    { 0x9a, 0x16, 0x0, 0x90, 0x27, 0x3f, 0xc1, 0x4d } \
+  }
+
+#define GRUB_EFI_ACPI_TABLE_GUID	\
+  { 0xeb9d2d30, 0x2d88, 0x11d3, \
+    { 0x9a, 0x16, 0x0, 0x90, 0x27, 0x3f, 0xc1, 0x4d } \
+  }
+
+#define GRUB_EFI_ACPI_20_TABLE_GUID	\
+  { 0x8868e871, 0xe4f1, 0x11d3, \
+    { 0xbc, 0x22, 0x0, 0x80, 0xc7, 0x3c, 0x88, 0x81 } \
+  }
+
+#define GRUB_EFI_SMBIOS_TABLE_GUID	\
+  { 0xeb9d2d31, 0x2d88, 0x11d3, \
+    { 0x9a, 0x16, 0x0, 0x90, 0x27, 0x3f, 0xc1, 0x4d } \
+  }
+
 /* Enumerations.  */
 enum grub_efi_timer_delay
   {
@@ -564,6 +583,16 @@ struct grub_efi_protocol_device_path
 };
 typedef struct grub_efi_protocol_device_path grub_efi_protocol_device_path_t;
 
+#define GRUB_EFI_PIWG_DEVICE_PATH_SUBTYPE		6
+
+struct grub_efi_piwg_device_path
+{
+  grub_efi_device_path_t header;
+  grub_efi_guid_t guid __attribute__ ((packed));
+};
+typedef struct grub_efi_piwg_device_path grub_efi_piwg_device_path_t;
+
+
 /* BIOS Boot Specification Device Path.  */
 #define GRUB_EFI_BIOS_DEVICE_PATH_TYPE			5
 
@@ -600,7 +629,7 @@ struct grub_efi_time
   grub_efi_int16_t time_zone;
   grub_efi_uint8_t daylight;
   grub_efi_uint8_t pad2;
-};
+} __attribute__ ((packed));
 typedef struct grub_efi_time grub_efi_time_t;
 
 struct grub_efi_time_capabilities
@@ -643,38 +672,38 @@ typedef struct grub_efi_table_header grub_efi_table_header_t;
 struct grub_efi_boot_services
 {
   grub_efi_table_header_t hdr;
-  
+
   grub_efi_tpl_t
   (*raise_tpl) (grub_efi_tpl_t new_tpl);
-  
+
   void
   (*restore_tpl) (grub_efi_tpl_t old_tpl);
-  
+
   grub_efi_status_t
   (*allocate_pages) (grub_efi_allocate_type_t type,
 		     grub_efi_memory_type_t memory_type,
 		     grub_efi_uintn_t pages,
 		     grub_efi_physical_address_t *memory);
-  
+
   grub_efi_status_t
   (*free_pages) (grub_efi_physical_address_t memory,
 		 grub_efi_uintn_t pages);
-  
+
   grub_efi_status_t
   (*get_memory_map) (grub_efi_uintn_t *memory_map_size,
 		     grub_efi_memory_descriptor_t *memory_map,
 		     grub_efi_uintn_t *map_key,
 		     grub_efi_uintn_t *descriptor_size,
 		     grub_efi_uint32_t *descriptor_version);
-  
+
   grub_efi_status_t
   (*allocate_pool) (grub_efi_memory_type_t pool_type,
 		    grub_efi_uintn_t size,
 		    void **buffer);
-  
+
   grub_efi_status_t
   (*free_pool) (void *buffer);
-  
+
   grub_efi_status_t
   (*create_event) (grub_efi_uint32_t type,
 		   grub_efi_tpl_t notify_tpl,
@@ -687,7 +716,7 @@ struct grub_efi_boot_services
   (*set_timer) (grub_efi_event_t event,
 		grub_efi_timer_delay_t type,
 		grub_efi_uint64_t trigger_time);
-  
+
    grub_efi_status_t
    (*wait_for_event) (grub_efi_uintn_t num_events,
 		      grub_efi_event_t *event,
@@ -695,10 +724,10 @@ struct grub_efi_boot_services
 
   grub_efi_status_t
   (*signal_event) (grub_efi_event_t event);
-  
+
   grub_efi_status_t
   (*close_event) (grub_efi_event_t event);
-  
+
   grub_efi_status_t
   (*check_event) (grub_efi_event_t event);
 
@@ -707,13 +736,13 @@ struct grub_efi_boot_services
 				  grub_efi_guid_t *protocol,
 				  grub_efi_interface_type_t interface_type,
 				  void *interface);
-  
+
   grub_efi_status_t
   (*reinstall_protocol_interface) (grub_efi_handle_t handle,
 				   grub_efi_guid_t *protocol,
 				   void *old_interface,
-				   void *new_inteface);
-  
+				   void *new_interface);
+
   grub_efi_status_t
   (*uninstall_protocol_interface) (grub_efi_handle_t handle,
 				   grub_efi_guid_t *protocol,
@@ -723,9 +752,9 @@ struct grub_efi_boot_services
   (*handle_protocol) (grub_efi_handle_t handle,
 		      grub_efi_guid_t *protocol,
 		      void **interface);
-  
+
   void *reserved;
-  
+
   grub_efi_status_t
   (*register_protocol_notify) (grub_efi_guid_t *protocol,
 			       grub_efi_event_t event,
@@ -855,7 +884,7 @@ struct grub_efi_runtime_services
 {
   grub_efi_table_header_t hdr;
 
-  grub_efi_status_t 
+  grub_efi_status_t
   (*get_time) (grub_efi_time_t *time,
 	       grub_efi_time_capabilities_t *capabilities);
 
@@ -914,19 +943,22 @@ struct grub_efi_configuration_table
 {
   grub_efi_guid_t vendor_guid;
   void *vendor_table;
-};
+} __attribute__ ((packed));
 typedef struct grub_efi_configuration_table grub_efi_configuration_table_t;
 
+#define GRUB_EFIEMU_SYSTEM_TABLE_SIGNATURE 0x5453595320494249LL
+#define GRUB_EFIEMU_RUNTIME_SERVICES_SIGNATURE 0x56524553544e5552LL
+
 struct grub_efi_simple_input_interface
 {
   grub_efi_status_t
   (*reset) (struct grub_efi_simple_input_interface *this,
 	    grub_efi_boolean_t extended_verification);
-  
+
   grub_efi_status_t
   (*read_key_stroke) (struct grub_efi_simple_input_interface *this,
 		      grub_efi_input_key_t *key);
-  
+
   grub_efi_event_t wait_for_key;
 };
 typedef struct grub_efi_simple_input_interface grub_efi_simple_input_interface_t;
@@ -951,7 +983,7 @@ struct grub_efi_simple_text_output_interface
 		 grub_efi_uintn_t *columns,
 		 grub_efi_uintn_t *rows);
 
-  grub_efi_status_t 
+  grub_efi_status_t
   (*set_mode) (struct grub_efi_simple_text_output_interface *this,
 	       grub_efi_uintn_t mode_number);
 
@@ -970,7 +1002,7 @@ struct grub_efi_simple_text_output_interface
   grub_efi_status_t
   (*enable_cursor) (struct grub_efi_simple_text_output_interface *this,
 		    grub_efi_boolean_t visible);
-  
+
   grub_efi_simple_text_output_mode_t *mode;
 };
 typedef struct grub_efi_simple_text_output_interface grub_efi_simple_text_output_interface_t;
@@ -1096,4 +1128,63 @@ struct grub_efi_block_io
 };
 typedef struct grub_efi_block_io grub_efi_block_io_t;
 
+#if GRUB_TARGET_SIZEOF_VOID_P == 4
+
+#define efi_call_0(func)		func()
+#define efi_call_1(func, a)		func(a)
+#define efi_call_2(func, a, b)		func(a, b)
+#define efi_call_3(func, a, b, c)	func(a, b, c)
+#define efi_call_4(func, a, b, c, d)	func(a, b, c, d)
+#define efi_call_5(func, a, b, c, d, e)	func(a, b, c, d, e)
+#define efi_call_6(func, a, b, c, d, e, f) func(a, b, c, d, e, f)
+#define efi_call_10(func, a, b, c, d, e, f, g, h, i, j)	func(a, b, c, d, e, f, g, h, i, j)
+
+#else
+
+#define efi_call_0(func) \
+  efi_wrap_0(func)
+#define efi_call_1(func, a) \
+  efi_wrap_1(func, (grub_uint64_t) a)
+#define efi_call_2(func, a, b) \
+  efi_wrap_2(func, (grub_uint64_t) a, (grub_uint64_t) b)
+#define efi_call_3(func, a, b, c) \
+  efi_wrap_3(func, (grub_uint64_t) a, (grub_uint64_t) b, (grub_uint64_t) c)
+#define efi_call_4(func, a, b, c, d) \
+  efi_wrap_4(func, (grub_uint64_t) a, (grub_uint64_t) b, (grub_uint64_t) c, \
+  (grub_uint64_t) d)
+#define efi_call_5(func, a, b, c, d, e)	\
+  efi_wrap_5(func, (grub_uint64_t) a, (grub_uint64_t) b, (grub_uint64_t) c, \
+  (grub_uint64_t) d, (grub_uint64_t) e)
+#define efi_call_6(func, a, b, c, d, e, f) \
+  efi_wrap_6(func, (grub_uint64_t) a, (grub_uint64_t) b, (grub_uint64_t) c, \
+  (grub_uint64_t) d, (grub_uint64_t) e, (grub_uint64_t) f)
+#define efi_call_10(func, a, b, c, d, e, f, g, h, i, j) \
+  efi_wrap_10(func, (grub_uint64_t) a, (grub_uint64_t) b, (grub_uint64_t) c, \
+  (grub_uint64_t) d, (grub_uint64_t) e, (grub_uint64_t) f, (grub_uint64_t) g, \
+  (grub_uint64_t) h, (grub_uint64_t) i, (grub_uint64_t) j)
+
+grub_uint64_t EXPORT_FUNC(efi_wrap_0) (void *func);
+grub_uint64_t EXPORT_FUNC(efi_wrap_1) (void *func, grub_uint64_t arg1);
+grub_uint64_t EXPORT_FUNC(efi_wrap_2) (void *func, grub_uint64_t arg1,
+                                       grub_uint64_t arg2);
+grub_uint64_t EXPORT_FUNC(efi_wrap_3) (void *func, grub_uint64_t arg1,
+                                       grub_uint64_t arg2, grub_uint64_t arg3);
+grub_uint64_t EXPORT_FUNC(efi_wrap_4) (void *func, grub_uint64_t arg1,
+                                       grub_uint64_t arg2, grub_uint64_t arg3,
+                                       grub_uint64_t arg4);
+grub_uint64_t EXPORT_FUNC(efi_wrap_5) (void *func, grub_uint64_t arg1,
+                                       grub_uint64_t arg2, grub_uint64_t arg3,
+                                       grub_uint64_t arg4, grub_uint64_t arg5);
+grub_uint64_t EXPORT_FUNC(efi_wrap_6) (void *func, grub_uint64_t arg1,
+                                       grub_uint64_t arg2, grub_uint64_t arg3,
+                                       grub_uint64_t arg4, grub_uint64_t arg5,
+                                       grub_uint64_t arg6);
+grub_uint64_t EXPORT_FUNC(efi_wrap_10) (void *func, grub_uint64_t arg1,
+                                        grub_uint64_t arg2, grub_uint64_t arg3,
+                                        grub_uint64_t arg4, grub_uint64_t arg5,
+                                        grub_uint64_t arg6, grub_uint64_t arg7,
+                                        grub_uint64_t arg8, grub_uint64_t arg9,
+                                        grub_uint64_t arg10);
+#endif
+
 #endif /* ! GRUB_EFI_API_HEADER */
